@@ -197,13 +197,12 @@ impl Einsum {
     #[apply(new_rc_unwrap)]
     pub fn try_new(args: EinsumArgs, out_axes: EinsumAxes, name: Option<String>) -> Result<Self> {
         // TODO: reduce axes nums!
-        let mut out = Self {
+        let out = Self {
             args,
             out_axes,
-            name: Default::default(),
+            name,
             info: Default::default(),
         };
-        out.name = out.auto_name(name);
 
         let map = out.shape_map()?; // catch errors
 
@@ -219,7 +218,7 @@ impl Einsum {
         }
 
         // could reuse hashmap if helped with perf
-        out.init_info()
+        out.initial_init_info()
     }
 
     pub fn try_from_spec(
@@ -251,17 +250,13 @@ impl Einsum {
 impl CircuitNodeAutoName for Einsum {
     const PRIORITY: OperatorPriority = OperatorPriority::Infix { priority: 1 };
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            children_names_with_maybe_paren(&Self::PRIORITY, self.children().collect()).map(
-                |names| {
-                    names
-                        .iter()
-                        .map(|name| Self::shorten_child_name(name))
-                        .collect::<Vec<String>>()
-                        .join(" * ")
-                },
-            )
+    fn auto_name(&self) -> Option<String> {
+        children_names_with_maybe_paren(&Self::PRIORITY, self.children().collect()).map(|names| {
+            names
+                .iter()
+                .map(|name| Self::shorten_child_name(name))
+                .collect::<Vec<String>>()
+                .join(" * ")
         })
     }
 }
@@ -551,15 +546,14 @@ impl Add {
     #[apply(new_rc_unwrap)]
     pub fn try_new(nodes: Vec<CircuitRc>, name: Option<String>) -> Result<Self> {
         // TODO: reduce axes nums!
-        let mut out = Self {
+        let out = Self {
             nodes,
-            name: Default::default(),
+            name,
             info: Default::default(),
         };
-        out.name = out.auto_name(name.clone());
         out.compute_shape_maybe()
-            .with_context(|| format!("Sum isn't broadcastable name={:?}", name))?;
-        out.init_info()
+            .with_context(|| format!("Sum isn't broadcastable name={:?}", out.name()))?;
+        out.initial_init_info()
     }
 
     pub fn try_from_counts(
@@ -718,17 +712,13 @@ impl CircuitNode for Add {
 impl CircuitNodeAutoName for Add {
     const PRIORITY: OperatorPriority = OperatorPriority::Infix { priority: 0 };
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            children_names_with_maybe_paren(&Self::PRIORITY, self.children().collect()).map(
-                |names| {
-                    names
-                        .iter()
-                        .map(|name| Self::shorten_child_name(name))
-                        .collect::<Vec<String>>()
-                        .join(" + ")
-                },
-            )
+    fn auto_name(&self) -> Option<String> {
+        children_names_with_maybe_paren(&Self::PRIORITY, self.children().collect()).map(|names| {
+            names
+                .iter()
+                .map(|name| Self::shorten_child_name(name))
+                .collect::<Vec<String>>()
+                .join(" + ")
         })
     }
 }
@@ -749,13 +739,12 @@ impl Rearrange {
     pub fn try_new(node: CircuitRc, spec: RearrangeSpec, name: Option<String>) -> Result<Self> {
         // spec must be valid because we enforce that on construction!
         // TODO: reduce axes nums!
-        let mut out = Self {
+        let out = Self {
             node: node.clone(),
             spec: Box::new(spec),
-            name: Default::default(),
+            name,
             info: Default::default(),
         };
-        out.name = out.auto_name(name.clone());
 
         // Check that RearrangeSpec is valid for input, prevent panic on rearrange.info().shape
         out.spec
@@ -764,7 +753,7 @@ impl Rearrange {
                 format!(
                     "{}\nname={:?}, child={:?}",
                     "rearrange spec conform_to_input_shape failed in construct rearrange",
-                    name,
+                    out.name(),
                     node
                 )
             })?;
@@ -802,7 +791,7 @@ impl Rearrange {
             }
             .into());
         }
-        out.init_info()
+        out.initial_init_info()
     }
 
     pub fn evolve(&self, node: Option<CircuitRc>, spec: Option<RearrangeSpec>) -> Rearrange {
@@ -887,39 +876,13 @@ impl CircuitNode for Rearrange {
             .to_einops_string_and_letter_sizes();
         einops_repeat(&tensors[0], string, letter_sizes)
     }
-
-    fn intermediate_cost_bound(&self) -> usize {
-        // Most rearranges are free, but some involve a copy
-        //
-        // Cases include:
-        // 1. Flattening axes out of order: a b -> (b a) is a copy unless either dimension is a singleton
-        // 2. Flattening a repeat: a -> (a 2) is a copy
-        // 3. Flattening a view with an offset/repeat:
-        //    * If x = ones(4,4) and y = x[:, 1:] and z = repeat(y, 'a b -> (a b)'), z is a copy
-        //    * If x = ones(4) and y = repeat(x, 'a -> 2 a') and z = repeat(y, 'a b -> (a b)'), z is a copy
-        //
-        // Because of the third case, we cannot confirm that a rearrange returns a view
-        // and has cost 0 unless we have the stride and offset of the child nodes
-        //
-        // Thus, for now, we always return the upper bound cost of numel. We'll explicitly
-        // track and propagate strides alongside shapes later to optimize this
-        if self.spec.output_ints.iter().any(|x| x.len() > 1) {
-            self.info().numel_usize()
-        } else {
-            // some flattens / repeats dont need copy, but we'd need stride to know that
-            0
-        }
-    }
 }
 
 impl CircuitNodeAutoName for Rearrange {
     const PRIORITY: OperatorPriority = OperatorPriority::PostFix {};
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            child_name_with_maybe_paren(&Self::PRIORITY, self.node.clone())
-                .map(|n| n + " rearrange")
-        })
+    fn auto_name(&self) -> Option<String> {
+        child_name_with_maybe_paren(&Self::PRIORITY, self.node.clone()).map(|n| n + " rearrange")
     }
 }
 
@@ -1022,14 +985,13 @@ impl Index {
 
         index.validate(&node.info().shape)?;
 
-        let mut out = Self {
+        let out = Self {
             node,
             index,
-            name: Default::default(),
+            name,
             info: Default::default(),
         };
-        out.name = out.auto_name(name);
-        out.init_info()
+        out.initial_init_info()
     }
 }
 
@@ -1107,31 +1069,13 @@ impl CircuitNode for Index {
                 .map(|x| x.extract(py).unwrap())
         })
     }
-
-    fn intermediate_cost_bound(&self) -> usize {
-        // Index returns a view if all of the axis indices are not int tensors
-        let has_tensor = self
-            .index
-            .0
-            .iter()
-            .any(|x| matches!(x, TensorAxisIndex::Tensor(_)));
-
-        if has_tensor {
-            // It might still return a view with int tensors in some cases, but we don't handle this
-            self.info().numel_usize()
-        } else {
-            0
-        }
-    }
 }
 
 impl CircuitNodeAutoName for Index {
     const PRIORITY: OperatorPriority = OperatorPriority::PostFix {};
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            child_name_with_maybe_paren(&Self::PRIORITY, self.node.clone()).map(|n| n + " idx")
-        })
+    fn auto_name(&self) -> Option<String> {
+        child_name_with_maybe_paren(&Self::PRIORITY, self.node.clone()).map(|n| n + " idx")
     }
 }
 
@@ -1353,14 +1297,13 @@ impl Concat {
             }
         }
 
-        let mut out = Self {
+        let out = Self {
             nodes,
             axis,
             info: Default::default(),
-            name: Default::default(),
+            name,
         };
-        out.name = out.auto_name(name);
-        out.init_info()
+        out.initial_init_info()
     }
 
     fn convert_axis(nodes: &[CircuitRc], axis: i64, is_stack: bool) -> Result<usize> {
@@ -1438,17 +1381,13 @@ impl CircuitNode for Concat {
 impl CircuitNodeAutoName for Concat {
     const PRIORITY: OperatorPriority = OperatorPriority::InfixAmbiguous {};
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            children_names_with_maybe_paren(&Self::PRIORITY, self.children().collect()).map(
-                |names| {
-                    names
-                        .iter()
-                        .map(|name| Self::shorten_child_name(name))
-                        .collect::<Vec<String>>()
-                        .join(" ++ ")
-                },
-            )
+    fn auto_name(&self) -> Option<String> {
+        children_names_with_maybe_paren(&Self::PRIORITY, self.children().collect()).map(|names| {
+            names
+                .iter()
+                .map(|name| Self::shorten_child_name(name))
+                .collect::<Vec<String>>()
+                .join(" ++ ")
         })
     }
 }
@@ -1544,13 +1483,12 @@ impl Scatter {
         let mut out = Self {
             node,
             index,
-            name: Default::default(),
+            name,
             info: Default::default(),
         };
-        out.name = out.auto_name(name);
         out.info.shape = shape;
         // todo check not too many axes / oob
-        out.init_info()
+        out.initial_init_info()
     }
 }
 
@@ -1620,19 +1558,13 @@ impl CircuitNode for Scatter {
                 .map(|x| x.extract(py).unwrap())
         })
     }
-
-    fn intermediate_cost_bound(&self) -> usize {
-        self.info().numel_usize()
-    }
 }
 
 impl CircuitNodeAutoName for Scatter {
     const PRIORITY: OperatorPriority = OperatorPriority::PostFix {};
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            child_name_with_maybe_paren(&Self::PRIORITY, self.node.clone()).map(|n| n + " scatter")
-        })
+    fn auto_name(&self) -> Option<String> {
+        child_name_with_maybe_paren(&Self::PRIORITY, self.node.clone()).map(|n| n + " scatter")
     }
 }
 
@@ -1745,13 +1677,12 @@ impl Conv {
             filter,
             stride,
             padding,
-            name: Default::default(),
+            name,
             info: Default::default(),
         };
         out.info.shape = out.try_compute_shape()?;
-        out.name = out.auto_name(name);
         // todo check not too many axes / oob
-        out.init_info()
+        out.initial_init_info()
     }
 
     pub fn try_compute_shape(&self) -> Result<Shape> {
@@ -1900,21 +1831,15 @@ impl CircuitNode for Conv {
             anyhow
         )
     }
-
-    fn intermediate_cost_bound(&self) -> usize {
-        self.info().numel_usize()
-    }
 }
 
 impl CircuitNodeAutoName for Conv {
     const PRIORITY: OperatorPriority = OperatorPriority::Function {};
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            let input_name = self.input.name()?;
-            let filter_name = self.filter.name()?;
-            Some(format!("Conv({input_name}, {filter_name})"))
-        })
+    fn auto_name(&self) -> Option<String> {
+        let input_name = self.input.name()?;
+        let filter_name = self.filter.name()?;
+        Some(format!("Conv({input_name}, {filter_name})"))
     }
 }
 
