@@ -20,7 +20,7 @@ use rr_util::{
     util::{arc_unwrap_or_clone, EinsumAxes, HashBytes},
 };
 use rustc_hash::FxHashMap as HashMap;
-use smallvec::SmallVec;
+use smallvec::{SmallVec, ToSmallVec};
 use thiserror::Error;
 
 use crate::{
@@ -203,7 +203,32 @@ pub fn expand_node(
     };
 
     let out = match &**circuit {
-        Circuit::SetSymbolicShape(_) => Ok(inputs[0].clone()), /* just return circuit, symbolic shapes will be added as needed! (TODO: maybe should keep?) */
+        Circuit::SetSymbolicShape(s) => {
+            let child_shape = inputs[0].shape();
+            let s2: Shape = s
+                .node
+                .shape()
+                .iter()
+                .zip(child_shape.iter().skip(batch_rank))
+                .zip(s.shape())
+                .map(|((old_child, new_child), us)| {
+                    if new_child == old_child {
+                        *us
+                    } else {
+                        *new_child
+                    }
+                })
+                .collect::<Shape>();
+            let s2 = [&(child_shape[..batch_rank]), s2.as_slice()]
+                .concat()
+                .to_smallvec();
+            let c = inputs[0].clone();
+            Ok(if s2 != *child_shape {
+                SetSymbolicShape::nrc(c, s2, s.name_cloned())
+            } else {
+                c
+            })
+        }
         Circuit::Array(_) | Circuit::Symbol(_) | Circuit::Scalar(_) => Ok(circuit.clone()),
         Circuit::Rearrange(node) => {
             let input_shape_non_batch = inputs[0].info().shape[batch_ranks[0]..]
