@@ -1,21 +1,24 @@
 use std::{
+    collections::BTreeSet,
     hash::{Hash, Hasher},
     iter::{once, zip},
     ops::Deref,
+    sync::Arc,
 };
 
 use anyhow::{bail, Context, Result};
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*, pyclass::CompareOp};
 use rr_util::{
-    cached_method,
+    cached_lambda, cached_method,
     caching::FastUnboundedCache,
-    py_types::use_rust_comp,
+    eq_by_big_hash, impl_eq_by_big_hash,
+    py_types::{use_rust_comp, PyCallable},
     pycall, python_error_exception,
     rearrange_spec::RearrangeSpec,
     symbolic_size::SymbolicSizeProduct,
     tensor_util::{right_align_shapes, Shape, TorchDeviceDtype},
-    util::{counts_g_1, is_unique},
+    util::{arc_unwrap_or_clone, counts_g_1, is_unique},
     IndexSet,
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet, FxHasher};
@@ -26,7 +29,7 @@ use crate::{
     circuit_node_auto_impl, circuit_node_extra_impl,
     circuit_node_private::{CircuitNodeHashItems, CircuitNodePrivate, CircuitNodeSetNonHashInfo},
     circuit_utils::OperatorPriority,
-    deep_map, deep_map_op_context_preorder_stoppable, deep_map_preorder_unwrap,
+    deep_map, deep_map_op, deep_map_op_context_preorder_stoppable, deep_map_preorder_unwrap,
     expand_node::{ExpandError, ReplaceExpander, ReplaceMapRc},
     named_axes::{deep_strip_axis_names, propagate_named_axes},
     new_rc_unwrap,
@@ -68,6 +71,57 @@ where
     out
 }
 
+#[derive(Clone, Debug)]
+pub struct SymbolSetRc {
+    set: Arc<BTreeSet<Symbol>>,
+    hash: HashBytes,
+}
+
+impl<'source> pyo3::FromPyObject<'source> for SymbolSetRc {
+    fn extract(from_py_obj: &'source pyo3::PyAny) -> pyo3::PyResult<Self> {
+        let from: BTreeSet<Symbol> = from_py_obj.extract()?;
+        Ok(Self::new(from))
+    }
+}
+
+impl Default for SymbolSetRc {
+    fn default() -> Self {
+        Self::new(Default::default())
+    }
+}
+
+impl Deref for SymbolSetRc {
+    type Target = BTreeSet<Symbol>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.set
+    }
+}
+
+impl SymbolSetRc {
+    pub fn new(set: BTreeSet<Symbol>) -> Self {
+        let mut hasher = blake3::Hasher::new();
+        for v in set.iter() {
+            hasher.update(&v.info().hash);
+        }
+        Self {
+            set: Arc::new(set),
+            hash: hasher.finalize().into(),
+        }
+    }
+
+    pub fn into_map(self) -> <Self as Deref>::Target {
+        arc_unwrap_or_clone(self.set)
+    }
+}
+
+impl eq_by_big_hash::EqByBigHash for SymbolSetRc {
+    fn hash(&self) -> HashBytes {
+        self.hash
+    }
+}
+impl_eq_by_big_hash!(SymbolSetRc);
+
 impl ModuleSpec {
     pub const EXPAND_PLACEHOLDER_UUID: Uuid = uuid!("741ba404-eec3-4ac9-b6ce-062e903fb033");
 
@@ -312,29 +366,41 @@ impl ModuleSpec {
         self.check_no_free_bound_by_nested(nodes)?;
 
         let shapes = nodes.iter().map(|x| x.info().shape.clone()).collect();
+        // TODO: maybe rep_hashes should really just be syms?
         let (expanded_shape, rep_hashes, clear_names_str) = self.expand_shape(&shapes)?;
         let node_mapping: HashMap<_, _> = rep_hashes.into_iter().zip(nodes).collect();
-        // we're replacing leaves so no need to do anything with stoppable preorder
-        let result = deep_map(expanded_shape, |x| {
+        // first update naming
+        let result = deep_map_op(expanded_shape.clone(), |x| {
+            // we only update naming if the given node has a bound free symbol
+            // (NOTE: these free symbols are now the *expanded* symbols from expand_shape)
+            if !x.is_symbol()
+                && x.info()
+                    .free_symbols
+                    .iter()
+                    .any(|x| node_mapping.contains_key(&x.info().hash))
+            {
+                if let Some(n) = x.name() {
+                    if n.contains(&clear_names_str) {
+                        let name = x.get_autoname().unwrap_or(None);
+                        return Some(x.rename(name));
+                    } else if name_prefix.is_some() || name_suffix.is_some() {
+                        let name = Some(
+                            name_prefix.as_deref().unwrap_or("").to_owned()
+                                + n
+                                + name_suffix.as_deref().unwrap_or(""),
+                        );
+                        return Some(x.rename(name));
+                    }
+                }
+            }
+            None
+        })
+        .unwrap_or(expanded_shape);
+        let result = deep_map(result, |x| {
             if let Some(r) = node_mapping.get(&x.info().hash) {
                 Ok((*r).clone())
             } else {
-                let mut y = x.clone();
-                if let Some(n) = x.name() {
-                    if n.contains(&clear_names_str) {
-                        y = y.rename(x.get_autoname().unwrap_or(None));
-                    }
-                    if name_prefix.is_some() || name_suffix.is_some() {
-                        if !node_mapping.contains_key(&x.info().hash) {
-                            y = y.rename(Some(
-                                name_prefix.clone().unwrap_or_else(|| "".to_owned())
-                                    + n
-                                    + name_suffix.as_ref().map(|x| -> &str { x }).unwrap_or(""),
-                            ));
-                        }
-                    }
-                }
-                Ok(y)
+                Ok(x)
             }
         })
         .with_context(|| {
@@ -388,6 +454,59 @@ impl ModuleSpec {
             arg_specs: self.arg_specs.clone(),
         }
     }
+
+    // we could probably make this faster...
+    pub fn map_on_replaced_path<F>(&self, f: F) -> Result<Self>
+    where
+        F: Fn(CircuitRc) -> Result<CircuitRc>,
+    {
+        #[apply(cached_lambda)]
+        #[key((circ.info().hash, symbols.hash), (HashBytes, HashBytes))]
+        #[use_try]
+        // Some iff some child bound
+        fn recurse(circ: CircuitRc, symbols: &SymbolSetRc) -> Result<Option<CircuitRc>> {
+            if let Some(sym) = circ.as_symbol() {
+                if symbols.contains(sym) {
+                    return Ok(Some(sym.crc()));
+                }
+            }
+            // TODO: maybe extract method...
+            let new = if let Some(m) = circ.as_module() {
+                let mut removed: BTreeSet<_> = symbols
+                    .iter()
+                    .filter(|x| m.info().free_symbols.contains(*x))
+                    .cloned()
+                    .collect();
+                for bound in &m.spec.arg_specs {
+                    removed.remove(&bound.symbol);
+                }
+
+                std::iter::once(recurse(m.spec.circuit.clone(), &SymbolSetRc::new(removed)))
+                    .chain(m.nodes.iter().map(|x| recurse(x.clone(), symbols)))
+                    .collect::<Result<Vec<_>>>()
+            } else {
+                circ.children().map(|x| recurse(x, symbols)).collect()
+            };
+            let new = new?;
+            assert_eq!(new.len(), circ.num_children());
+            // note that bound symbol children are always returned as Some (see above)
+            if new.iter().all(|x| x.is_none()) {
+                return Ok(None);
+            }
+            let new_circ = circ.map_children_unwrap_enumerate(|i, c| new[i].clone().unwrap_or(c));
+
+            f(new_circ).map(Some)
+        }
+        let symbols = SymbolSetRc::new(self.arg_specs.iter().map(|x| x.symbol.clone()).collect());
+        self.map_circuit(|circ| recurse(circ.clone(), &symbols).map(|x| x.unwrap_or(circ)))
+    }
+
+    pub fn map_on_replaced_path_unwrap<F>(&self, f: F) -> Self
+    where
+        F: Fn(CircuitRc) -> CircuitRc,
+    {
+        self.map_on_replaced_path(|x| Ok(f(x))).unwrap()
+    }
 }
 
 pub enum FreeSymbolsWrapper<'a> {
@@ -605,6 +724,30 @@ impl ModuleSpec {
         self.hash(&mut s);
         s.finish()
     }
+
+    #[pyo3(name = "map_on_replaced_path")]
+    pub fn py_map_on_replaced_path(&self, f: PyCallable) -> Result<Self> {
+        self.map_on_replaced_path(|x| pycall!(f, (x,), anyhow))
+    }
+
+    pub fn rename_on_replaced_path(&self, prefix: Option<String>, suffix: Option<String>) -> Self {
+        if prefix.is_none() && suffix.is_none() {
+            return self.clone();
+        }
+
+        self.map_on_replaced_path_unwrap(|x| {
+            if let Some(n) = x.name() {
+                let name = Some(
+                    prefix.as_deref().unwrap_or("").to_owned()
+                        + n
+                        + suffix.as_deref().unwrap_or(""),
+                );
+                x.rename(name)
+            } else {
+                x
+            }
+        })
+    }
 }
 
 #[pyclass]
@@ -987,23 +1130,33 @@ impl Module {
         name_suffix: Option<String>,
         use_self_name_as_prefix: bool,
     ) -> Result<CircuitRc> {
-        let name_prefix = if use_self_name_as_prefix {
-            if let Some(name_prefix) = name_prefix {
-                bail!(
-                    ConstructError::ModulePassedNamePrefixAndUseSelfNameAsPrefix {
-                        name_prefix,
-                        module: self.clone()
-                    }
-                )
-            }
-
-            self.name_cloned().map(|x| x + ".")
-        } else {
-            name_prefix
-        };
+        let name_prefix = self.get_prefix(name_prefix, use_self_name_as_prefix)?;
         Ok(self.substitute(name_prefix, name_suffix))
     }
 
+    #[args(prefix = "None", suffix = "None", use_self_name_as_prefix = "false")]
+    pub fn rename_on_replaced_path(
+        &self,
+        prefix: Option<String>,
+        suffix: Option<String>,
+        use_self_name_as_prefix: bool,
+    ) -> Result<Module> {
+        let prefix = self.get_prefix(prefix, use_self_name_as_prefix)?;
+        Ok(Self::new(
+            self.nodes.clone(),
+            self.spec.rename_on_replaced_path(prefix, suffix),
+            self.name_cloned(),
+        ))
+    }
+
+    pub fn map_on_replaced_path(&self, f: PyCallable) -> Result<Self> {
+        Ok(Self::new(
+            self.nodes.clone(),
+            self.spec.py_map_on_replaced_path(f)?,
+            self.name_cloned(),
+        ))
+    }
+
     pub fn aligned_batch_shape(&self) -> Shape {
         self.spec.aligned_batch_shape(&self.nodes).unwrap()
     }
@@ -1095,6 +1248,27 @@ impl Module {
             .expect("constructor should ensure this works")
             .rename(self.name_cloned())
     }
+
+    fn get_prefix(
+        &self,
+        name_prefix: Option<String>,
+        use_self_name_as_prefix: bool,
+    ) -> Result<Option<String>> {
+        if use_self_name_as_prefix {
+            if let Some(name_prefix) = name_prefix {
+                bail!(
+                    ConstructError::ModulePassedNamePrefixAndUseSelfNameAsPrefix {
+                        name_prefix,
+                        module: self.clone()
+                    }
+                )
+            }
+
+            Ok(self.name_cloned().map(|x| x + "."))
+        } else {
+            Ok(name_prefix)
+        }
+    }
 }
 
 #[derive(Default)]
@@ -1177,12 +1351,15 @@ pub fn has_free_sym(circuit: &CircuitRc, sym: &Symbol) -> bool {
     }
 }
 
-pub fn is_intersecting_free_syms_slice(circuit: &CircuitRc, syms: &[&Symbol]) -> bool {
+pub fn is_intersecting_free_syms_iter<'a>(
+    circuit: &CircuitRc,
+    syms: impl IntoIterator<Item = &'a Symbol>,
+) -> bool {
     if let Some(other) = circuit.as_symbol() {
-        syms.contains(&other)
+        syms.into_iter().any(|x| x == other)
     } else {
-        syms.iter()
-            .any(|x| circuit.info().free_symbols.contains(*x))
+        syms.into_iter()
+            .any(|x| circuit.info().free_symbols.contains(x))
     }
 }
 
@@ -1203,7 +1380,7 @@ pub fn recur_check_no_free_bound_by_nested(
     if seen.insert(circ.info().hash) {
         if let Some(m) = circ.as_module() {
             // first check if we can term early (maybe not actually optimization)
-            if !is_intersecting_free_syms_slice(&circ, syms) {
+            if !is_intersecting_free_syms_iter(&circ, syms.iter().map(|x| *x)) {
                 return Ok(());
             }
 
@@ -1221,7 +1398,7 @@ pub fn recur_check_no_free_bound_by_nested(
                 })
                 .unzip();
 
-            if is_intersecting_free_syms_slice(&m.spec.circuit, &syms) {
+            if is_intersecting_free_syms_iter(&m.spec.circuit, syms.iter().map(|x| *x)) {
                 // fail earlier if possible
                 recur_check_no_free_bound_by_nested(m.spec.circuit.clone(), &syms, &nodes, seen)?;
 
