@@ -17,7 +17,7 @@ use rr_util::{
     timed,
     util::{counts_g_1, flip_op_result, is_unique, NamedAxes},
 };
-use rustc_hash::FxHashMap as HashMap;
+use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
 use uuid::Uuid;
 
@@ -563,7 +563,13 @@ impl Parser {
         top_level
             .iter()
             .map(|ident| {
-                self.deep_convert_partial_circ(ident, partial_circuits, context, tensor_cache)
+                self.deep_convert_partial_circ(
+                    ident,
+                    partial_circuits,
+                    context,
+                    tensor_cache,
+                    &mut Default::default(),
+                )
             })
             .collect()
     }
@@ -688,6 +694,9 @@ impl Parser {
         partial_circuits: &HashMap<CircuitIdent, PartialCirc>,
         context: &mut HashMap<CircuitIdent, CircuitRc>,
         tensor_cache: &mut Option<TensorCacheRrfs>,
+        in_progress: &mut HashSet<CircuitIdent>,
+        // mutable set of ancestor idents, insert when we start and remove when we finish
+        // to detect cycles (when something is an ancestor of itself)
     ) -> Result<CircuitRc> {
         if let Some(ref_circ) = ident
             .as_str()
@@ -699,11 +708,24 @@ impl Parser {
         if let Some(already) = context.get(ident) {
             return Ok(already.clone());
         }
+        if !in_progress.insert(ident.clone()) {
+            bail!(ParseCircuitError::Cycle {
+                ident: ident.clone()
+            });
+        }
         let ps = &partial_circuits[ident];
         let children: Vec<CircuitRc> = ps
             .children
             .iter()
-            .map(|x| self.deep_convert_partial_circ(x, partial_circuits, context, tensor_cache))
+            .map(|x| {
+                self.deep_convert_partial_circ(
+                    x,
+                    partial_circuits,
+                    context,
+                    tensor_cache,
+                    in_progress,
+                )
+            })
             .collect::<Result<Vec<_>, _>>()?;
 
         let result = self
@@ -720,6 +742,7 @@ impl Parser {
             .rename(partial_circuits[ident].name.clone());
 
         context.insert(ident.clone(), result.clone());
+        in_progress.remove(ident);
         Ok(result)
     }
 
@@ -1086,6 +1109,9 @@ pub enum ParseCircuitError {
     #[error("extra='{extra}' for variant={variant}, but expected empty ({e_name})")]
     ExpectedNoExtraInfo { extra: String, variant: CircuitType },
 
+    #[error("Trying to parse cycle! identifier {ident:?} ({e_name})")]
+    Cycle { ident: CircuitIdent },
+
     #[error("Parsing wrong number of children, expected {expected} found {found}, ident={ident:?} ({e_name})")]
     WrongNumberChildren {
         expected: usize,
