@@ -1,7 +1,6 @@
-use std::collections::HashMap;
-
 use circuit_base::{deep_map_op, deep_map_op_context, prelude::*, Add, Einsum, Index, Rearrange};
 use pyo3::prelude::*;
+use rustc_hash::FxHashMap as HashMap;
 
 use crate::circuit_optimizer::OptimizationContext;
 
@@ -67,7 +66,7 @@ pub fn deep_canonicalize(circuit: CircuitRc, context: &mut OptimizationContext)
     deep_map_op_context(
         circuit.clone(),
         &|x, _c: &mut HashMap<(), ()>| canonicalize_node_op(x),
-        &mut HashMap::<(), ()>::new(),
+        &mut HashMap::<(), ()>::default(),
         &mut context.cache.canonicalized,
     )
     .unwrap_or(circuit)
