@@ -1,14 +1,13 @@
 use std::{
-    collections::{btree_map::Entry, BTreeMap, BTreeSet},
+    collections::{btree_map::Entry, BTreeMap},
     iter,
-    ops::Deref,
     sync::Arc,
 };
 
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
     expand_node::{expand_node, MapReplaceExpander, ReplaceMapRc},
-    module::{are_args_used, is_intersecting_free_syms},
+    module::{are_args_used, is_intersecting_free_syms, SymbolSetRc},
     CircuitNode, CircuitRc, Module, ModuleArgSpec, ModuleSpec, Rearrange, Symbol,
 };
 use circuit_rewrites::{
@@ -30,7 +29,7 @@ use rr_util::{
     fn_struct, impl_eq_by_big_hash, pycall, python_error_exception,
     rearrange_spec::{OpSize, RearrangeSpec, RearrangeSpecError},
     sv, tu8v, unwrap,
-    util::{arc_unwrap_or_clone, DimNumMaker, HashBytes},
+    util::{DimNumMaker, HashBytes},
     IndexSet,
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
@@ -1325,57 +1324,6 @@ pub fn update_bindings_nested(
     .map(|x| unwrap!(x, CircOrGet::Circ))
 }
 
-#[derive(Clone, Debug)]
-pub struct SymbolSetRc {
-    set: Arc<BTreeSet<Symbol>>,
-    hash: HashBytes,
-}
-
-impl<'source> pyo3::FromPyObject<'source> for SymbolSetRc {
-    fn extract(from_py_obj: &'source pyo3::PyAny) -> pyo3::PyResult<Self> {
-        let from: BTreeSet<Symbol> = from_py_obj.extract()?;
-        Ok(Self::new(from))
-    }
-}
-
-impl Default for SymbolSetRc {
-    fn default() -> Self {
-        Self::new(Default::default())
-    }
-}
-
-impl Deref for SymbolSetRc {
-    type Target = BTreeSet<Symbol>;
-
-    fn deref(&self) -> &Self::Target {
-        &self.set
-    }
-}
-
-impl SymbolSetRc {
-    pub fn new(set: BTreeSet<Symbol>) -> Self {
-        let mut hasher = blake3::Hasher::new();
-        for v in set.iter() {
-            hasher.update(&v.hash());
-        }
-        Self {
-            set: Arc::new(set),
-            hash: hasher.finalize().into(),
-        }
-    }
-
-    pub fn into_map(self) -> <Self as Deref>::Target {
-        arc_unwrap_or_clone(self.set)
-    }
-}
-
-impl EqByBigHash for SymbolSetRc {
-    fn hash(&self) -> HashBytes {
-        self.hash
-    }
-}
-impl_eq_by_big_hash!(SymbolSetRc);
-
 struct SymbolExtractor {
     cache: FastUnboundedCache<
         (HashBytes, HashBytes, IterativeMatcherRc),
